"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.websqlError = exports.getSize = exports.compactRevs = exports.select = exports.qMarks = exports.unstringifyDoc = exports.stringifyDoc = exports.unescapeBlob = exports.escapeBlob = void 0;
const pouchdb_errors_1 = require("pouchdb-errors");
const pouchdb_utils_1 = require("pouchdb-utils");
const constants_1 = require("./constants");
function escapeBlob(str) {
    return str
        .replace(/\u0002/g, '\u0002\u0002')
        .replace(/\u0001/g, '\u0001\u0002')
        .replace(/\u0000/g, '\u0001\u0001');
}
exports.escapeBlob = escapeBlob;
function unescapeBlob(str) {
    return str
        .replace(/\u0001\u0001/g, '\u0000')
        .replace(/\u0001\u0002/g, '\u0001')
        .replace(/\u0002\u0002/g, '\u0002');
}
exports.unescapeBlob = unescapeBlob;
function stringifyDoc(doc) {
    delete doc._id;
    delete doc._rev;
    return JSON.stringify(doc);
}
exports.stringifyDoc = stringifyDoc;
function unstringifyDoc(doc, id, rev) {
    const it = JSON.parse(doc);
    it._id = id;
    it._rev = rev;
    return it;
}
exports.unstringifyDoc = unstringifyDoc;
function qMarks(num) {
    let s = '(';
    while (num--) {
        s += '?';
        if (num) {
            s += ',';
        }
    }
    return s + ')';
}
exports.qMarks = qMarks;
function select(selector, table, joiner, where, orderBy) {
    return ('SELECT ' +
        selector +
        ' FROM ' +
        (typeof table === 'string' ? table : table.join(' JOIN ')) +
        (joiner ? ' ON ' + joiner : '') +
        (where ? ' WHERE ' + (typeof where === 'string' ? where : where.join(' AND ')) : '') +
        (orderBy ? ' ORDER BY ' + orderBy : ''));
}
exports.select = select;
function compactRevs(revs, docId, tx) {
    if (!revs.length) {
        return;
    }
    let numDone = 0;
    const seqs = [];
    function checkDone() {
        if (++numDone === revs.length) {
            deleteOrphans();
        }
    }
    function deleteOrphans() {
        if (!seqs.length) {
            return;
        }
        const sql = 'SELECT DISTINCT digest AS digest FROM ' +
            constants_1.ATTACH_AND_SEQ_STORE +
            ' WHERE seq IN ' +
            qMarks(seqs.length);
        tx.executeSql(sql, seqs, function (tx, res) {
            const digestsToCheck = [];
            for (let i = 0; i < res.rows.length; i++) {
                digestsToCheck.push(res.rows.item(i).digest);
            }
            if (!digestsToCheck.length) {
                return;
            }
            const sql = 'DELETE FROM ' +
                constants_1.ATTACH_AND_SEQ_STORE +
                ' WHERE seq IN (' +
                seqs
                    .map(function () {
                    return '?';
                })
                    .join(',') +
                ')';
            tx.executeSql(sql, seqs, function (tx) {
                const sql = 'SELECT digest FROM ' +
                    constants_1.ATTACH_AND_SEQ_STORE +
                    ' WHERE digest IN (' +
                    digestsToCheck
                        .map(function () {
                        return '?';
                    })
                        .join(',') +
                    ')';
                tx.executeSql(sql, digestsToCheck, function (tx, res) {
                    const nonOrphanedDigests = new Set();
                    for (let i = 0; i < res.rows.length; i++) {
                        nonOrphanedDigests.add(res.rows.item(i).digest);
                    }
                    digestsToCheck.forEach(function (digest) {
                        if (nonOrphanedDigests.has(digest)) {
                            return;
                        }
                        tx.executeSql('DELETE FROM ' + constants_1.ATTACH_AND_SEQ_STORE + ' WHERE digest=?', [digest]);
                        tx.executeSql('DELETE FROM ' + constants_1.ATTACH_STORE + ' WHERE digest=?', [digest]);
                    });
                });
            });
        });
    }
    revs.forEach(function forEachRevision(rev) {
        const sql = 'SELECT seq FROM ' + constants_1.BY_SEQ_STORE + ' WHERE doc_id=? AND rev=?';
        tx.executeSql(sql, [docId, rev], function (tx, res) {
            try {
                if (!res.rows.length) {
                    return checkDone();
                }
                const firstRow = res.rows.item(0);
                if (firstRow !== undefined) {
                    seqs.push(firstRow.seq);
                    tx.executeSql('DELETE FROM ' + constants_1.BY_SEQ_STORE + ' WHERE seq=?', [firstRow.seq], checkDone);
                }
                else {
                    const args = {
                        type: 'ROW_IS_UNDEFINED',
                        sql,
                        rows: res.rows,
                        rows_length: res.rows.length,
                    };
                    console.log('compactRevs::deleteOrphans[error: ROW_IS_UNDEFINED] -> ', args);
                }
            }
            catch (err) {
                const args = {
                    type: 'error',
                    sql,
                    rows: res.rows,
                    rows_length: res.rows.length,
                };
                console.log('compactRevs::deleteOrphans[error: ' + err.message || 'unknown' + '] -> ', args);
                console.error(err);
            }
        });
    });
}
exports.compactRevs = compactRevs;
function websqlError(callback) {
    return function (event) {
        pouchdb_utils_1.guardedConsole('error', 'WebSQL threw an error', event);
        const errorNameMatch = event && event.constructor.toString().match(/function ([^(]+)/);
        const errorName = (errorNameMatch && errorNameMatch[1]) || event.type;
        const errorReason = event.target || event.message;
        callback(pouchdb_errors_1.createError(pouchdb_errors_1.WSQ_ERROR, errorReason, errorName));
    };
}
exports.websqlError = websqlError;
function getSize(opts) {
    if ('size' in opts) {
        return opts.size * 1000000;
    }
    const isAndroid = typeof navigator !== 'undefined' && /Android/.test(navigator.userAgent);
    return isAndroid ? 5000000 : 1;
}
exports.getSize = getSize;
//# sourceMappingURL=utils.js.map